import gzip
import json

from pyrsistent import freeze

from lexrank import settings

file = settings.ASSETS_ROOT / 'stopwords.json.gz'

with gzip.open(file, mode='rt', encoding='utf-8') as fp:
    _STOPWORDS = json.load(fp)

STOPWORDS = {}

for lang, stopwords in _STOPWORDS.items():
    STOPWORDS[lang] = set(stopwords)

STOPWORDS = freeze(STOPWORDS)



import regex

EMAIL_REGEX = regex.compile(
    r'[\p{L}0-9]+[\p{L}0-9_.+-]*[\p{L}0-9_+-]+@[\p{L}0-9]+[\p{L}0-9.-]*\.\p{L}+'  # noqa
)
PUNCTUATION_SIGNS = set('.,;:¡!¿?…⋯&‹›«»\"“”[]()⟨⟩}{/|\\')


def clean_text(text, allowed_chars='- '):
    text = ' '.join(text.lower().split())
    text = ''.join(ch for ch in text if ch.isalnum() or ch in allowed_chars)

    return text


def contains_letters(word):
    return any(ch.isalpha() for ch in word)


def contains_numbers(word):
    return any(ch.isdigit() for ch in word)


def filter_words(words, stopwords, keep_numbers=False):
    if keep_numbers:
        words = [
            word for word in words
            if (contains_letters(word) or contains_numbers(word)) and
            word not in stopwords
        ]

    else:
        words = [
            word for word in words
            if contains_letters(word) and not contains_numbers(word) and
            word not in stopwords
        ]

    return words


def separate_punctuation(text):
    text_punctuation = set(text) & PUNCTUATION_SIGNS

    for ch in text_punctuation:
        text = text.replace(ch, ' ' + ch + ' ')

    return text


def tokenize(text, stopwords, keep_numbers=False, keep_emails=False):
    tokens = []
    ix = 0

    for found in EMAIL_REGEX.finditer(text):
        part = text[ix: found.start()]
        words = clean_text(separate_punctuation(part)).split()
        words = filter_words(words, stopwords, keep_numbers=keep_numbers)

        email = found.group()
        ix = found.end()

        tokens.extend(words)

        if keep_emails:
            tokens.append(email)

    words = clean_text(separate_punctuation(text[ix:])).split()
    words = filter_words(words, stopwords, keep_numbers=keep_numbers)
    tokens.extend(words)

    return tokens



import numpy as np
from scipy.sparse.csgraph import connected_components


def _power_method(transition_matrix, increase_power=True):
    eigenvector = np.ones(len(transition_matrix))

    if len(eigenvector) == 1:
        return eigenvector

    transition = transition_matrix.transpose()

    while True:
        eigenvector_next = np.dot(transition, eigenvector)

        if np.allclose(eigenvector_next, eigenvector):
            return eigenvector_next

        eigenvector = eigenvector_next

        if increase_power:
            transition = np.dot(transition, transition)


def connected_nodes(matrix):
    _, labels = connected_components(matrix)
    groups = {}

    for ix, label in enumerate(labels):
        if label in groups:
            groups[label].append(ix)
            continue

        groups[label] = [ix]

    return list(groups.values())


def stationary_distribution(transition_matrix, increase_power=True):
    num_nodes = len(transition_matrix)
    grouped_indices = connected_nodes(transition_matrix)

    eigenvector = []
    indices = []

    for group in grouped_indices:
        t_matrix = transition_matrix[np.ix_(group, group)]
        eigenvector.extend(
            _power_method(t_matrix, increase_power=increase_power),
        )
        indices.extend(group)

    return [float(eigenvector[ix]) / num_nodes for ix in indices]



from pathlib import Path

from setuptools import find_packages, setup

requirements = Path(__file__).parent / 'requirements/core.txt'

with requirements.open(mode='rt', encoding='utf-8') as fp:
    install_requires = [line.strip() for line in fp]

readme = Path(__file__).parent / 'README.rst'

with readme.open(mode='rt', encoding='utf-8') as fp:
    readme_text = fp.read()

setup(
    name='lexrank',
    maintainer='Ocean S.A.',
    maintainer_email='support@ocean.io',
    version='0.0.1b',
    description='LexRank text summarization',
    long_description=readme_text,
    keywords=[
        'lex', 'rank', 'lexrank', 'algorithm', 'text', 'summary',
        'summarization',
    ],
    license='MIT',
    author='Luka Shostenko',
    author_email='luka.shostenko@gmail.com',
    url='https://github.com/wikibusiness/lexrank',
    download_url='https://github.com/wikibusiness/lexrank/archive/0.0.1a.tar.gz',  # noqa
    packages=find_packages(include=['lexrank.*']),
    py_modules=['lexrank.settings'],
    python_requires='>=3.5.0',
    install_requires=install_requires,
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'assemble_stopwords = lexrank.tools.assemble_stopwords:entrypoint',
        ],
    },
)



import math
from collections import defaultdict

import numpy as np

from lexrank.algorithms.power_method import stationary_distribution
from lexrank.utils.text import tokenize


class LexRank:
    def __init__(
        self,
        documents,
        stopwords=None,
        keep_numbers=False,
        keep_emails=False,
        include_new_words=True,
    ):
        if stopwords is None:
            self.stopwords = set()
        else:
            self.stopwords = stopwords

        self.keep_numbers = keep_numbers
        self.keep_emails = keep_emails
        self.include_new_words = include_new_words

        bags_of_words = []

        for doc in documents:
            doc_words = set()

            for sentence in doc:
                words = self.tokenize_sentence(sentence)
                doc_words.update(words)

            if doc_words:
                bags_of_words.append(doc_words)

        if not bags_of_words:
            raise ValueError('documents are not informative')

        self.bags_of_words = bags_of_words
        self.idf_score = self._calculate_idf()

    def tokenize_sentence(self, sentence):
        tokens = tokenize(
            sentence,
            self.stopwords,
            keep_numbers=self.keep_numbers,
            keep_emails=self.keep_emails,
        )

        return tokens

    def _calculate_idf(self):
        doc_number_total = len(self.bags_of_words)

        if self.include_new_words:
            default_value = math.log(doc_number_total + 1)

        else:
            default_value = 0

        idf_score = defaultdict(lambda: default_value)

        for word in set.union(*self.bags_of_words):
            doc_number_word = sum(1 for bag in self.bags_of_words if word in bag)  # noqa
            idf_score[word] = math.log(doc_number_total / doc_number_word)

        return idf_score

    def _calculate_tf(self, tokenized_sentence):
        tf_score = {}

        for word in set(tokenized_sentence):
            tf = tokenized_sentence.count(word)
            tf_score[word] = tf

        return tf_score

    def _idf_modified_cosine(self, tf_scores, i, j):
        if i == j:
            return 1

        tf_i, tf_j = tf_scores[i], tf_scores[j]
        words_i, words_j = set(tf_i.keys()), set(tf_j.keys())

        nominator = 0

        for word in words_i & words_j:
            idf = self.idf_score[word]
            nominator += tf_i[word] * tf_j[word] * idf ** 2

        if math.isclose(nominator, 0):
            return 0

        denominator_i, denominator_j = 0, 0

        for word in words_i:
            tfidf = tf_i[word] * self.idf_score[word]
            denominator_i += tfidf ** 2

        for word in words_j:
            tfidf = tf_j[word] * self.idf_score[word]
            denominator_j += tfidf ** 2

        similarity = nominator / math.sqrt(denominator_i * denominator_j)

        return similarity

    def _calculate_similarity_matrix(self, tf_scores):
        length = len(tf_scores)

        similarity_matrix = np.zeros([length] * 2)

        for i in range(length):
            for j in range(i, length):
                similarity = self._idf_modified_cosine(tf_scores, i, j)

                if similarity:
                    similarity_matrix[i, j] = similarity
                    similarity_matrix[j, i] = similarity

        return similarity_matrix

    def _markov_matrix_discrete(self, similarity_matrix, threshold):
        markov_matrix = np.zeros(similarity_matrix.shape)

        for i in range(len(similarity_matrix)):
            columns = np.where(similarity_matrix[i] > threshold)[0]  # noqa
            markov_matrix[i, columns] = 1 / len(columns)

        return markov_matrix

    def _markov_matrix(self, similarity_matrix):
        row_sum = similarity_matrix.sum(axis=1, keepdims=True)

        return similarity_matrix / row_sum

    def rank_sentences(
        self,
        sentences,
        threshold=.03,
        discretize=True,
        fast_power_method=True,
        normalize=False,
    ):
        if not isinstance(threshold, float) or not 0 <= threshold < 1:
            raise ValueError(
                '\'threshold\' should be a floating-point number '
                'from the interval [0, 1)',
            )

        tf_scores = [
            self._calculate_tf(self.tokenize_sentence(sentence))
            for sentence in sentences
        ]

        similarity_matrix = self._calculate_similarity_matrix(tf_scores)

        if discretize:
            markov_matrix = self._markov_matrix_discrete(
                similarity_matrix,
                threshold=threshold,
            )

        else:
            markov_matrix = self._markov_matrix(similarity_matrix)

        lexrank = stationary_distribution(
            markov_matrix,
            increase_power=fast_power_method,
        )

        if normalize:
            max_val = max(lexrank)
            lexrank = [val / max_val for val in lexrank]

        return lexrank

    def get_summary(
        self,
        sentences,
        summary_size=1,
        threshold=.03,
        discretize=True,
        fast_power_method=True,
    ):
        if not isinstance(summary_size, int) or summary_size < 1:
            raise ValueError('\'summary_size\' should be a positive integer')

        lexrank = self.rank_sentences(
            sentences,
            threshold=threshold,
            discretize=discretize,
            fast_power_method=fast_power_method,
        )

        sorted_ix = np.argsort(lexrank)[::-1]
        summary = [sentences[i] for i in sorted_ix[:summary_size]]

        return summary



import argparse
import gzip
import json
import sys

from path import Path

from lexrank import settings
from lexrank.utils.text import clean_text


def assemble_stopwords(source_dir):
    files = source_dir.files('*.txt')
    stopwords_map = {}

    for file_path in files:
        lang = file_path.namebase
        lang_words = set()

        with file_path.open(mode='rt', encoding='utf-8') as fp:
            for name in fp.readlines():
                lang_words.add(clean_text(name))

        stopwords_map[lang] = list(lang_words)
        sys.stdout.write('  ' + lang + ' ' + 'stopwords collected\n')

    if not stopwords_map:
        return

    out_file = settings.ASSETS_ROOT / 'stopwords.json.gz'

    with gzip.open(out_file, mode='wt', encoding='utf-8') as fp:
        json.dump(stopwords_map, fp)

        msg = 'Stopwords written to {out_file}\n'.format(out_file=out_file)
        sys.stdout.write(msg)


def setup_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--source_dir', type=Path, required=True)

    return parser


def entrypoint():
    parser = setup_parser()
    options = parser.parse_args(sys.argv[1:])

    assemble_stopwords(options.source_dir)
